const { cupcake } = require('@lib/proto');
const FilesystemMap = require('@lib/fs_map');
const DataCompressor = require('@lib/compression');
const { deepClone } = require('@lib/proto');

const compressor = new DataCompressor();

//const global_cupkake_champions = {}; // Cache for champions by language key
const global_championCache = {}; // Cache for champions (To improve performance by changing data model)

let champion_en_US = null;

const setCacheDriver = () => {
    if (process.env.CACHE_DRIVER === "FS") {
        return new FilesystemMap();
    }

    return new Map();
}

const imageCache = setCacheDriver();
let served_bytes = 0;

const skinlist = { 266: 7, 103: 76, 84: 71, 166: 10, 12: 20, 32: 34, 34: 27, 1: 22, 523: 30, 22: 65, 136: 21, 268: 14, 432: 17, 200: 10, 53: 56, 63: 8, 201: 10, 233: 1, 51: 39, 164: 32, 69: 9, 31: 7, 42: 6, 122: 33, 131: 26, 119: 29, 36: 21, 245: 19, 60: 24, 28: 42, 81: 33, 9: 27, 114: 69, 105: 26, 3: 6, 41: 23, 86: 14, 150: 22, 79: 20, 104: 35, 887: 20, 120: 31, 74: 24, 910: 1, 420: 18, 39: 37, 427: 11, 40: 46, 59: 7, 24: 13, 126: 5, 202: 36, 222: 51, 145: 29, 429: 14, 43: 54, 30: 17, 38: 6, 55: 48, 10: 57, 141: 20, 85: 25, 121: 69, 203: 3, 240: 9, 96: 19, 897: 8, 7: 20, 64: 41, 89: 50, 876: 19, 127: 33, 236: 52, 117: 46, 99: 38, 54: 27, 90: 38, 57: 7, 11: 42, 902: 1, 21: 41, 62: 6, 82: 44, 25: 60, 950: 1, 267: 51, 75: 25, 111: 9, 518: 31, 76: 9, 895: 11, 56: 7, 20: 26, 2: 25, 61: 20, 516: 11, 80: 36, 78: 33, 555: 53, 246: 40, 133: 14, 497: 5, 33: 17, 421: 17, 526: 10, 888: 21, 58: 42, 107: 30, 92: 23, 68: 23, 13: 20, 360: 30, 113: 7, 235: 56, 147: 24, 875: 56, 35: 43, 98: 49, 102: 17, 27: 19, 14: 30, 15: 51, 72: 5, 37: 46, 16: 27, 50: 21, 517: 36, 134: 44, 223: 30, 163: 21, 91: 39, 44: 2, 17: 18, 412: 39, 18: 51, 48: 6, 23: 5, 4: 25, 29: 36, 77: 2, 6: 15, 110: 44, 67: 44, 45: 51, 161: 20, 711: 10, 254: 30, 234: 30, 112: 5, 8: 30, 106: 9, 19: 35, 498: 4, 101: 5, 5: 27, 157: 57, 777: 45, 83: 30, 350: 1, 154: 14, 238: 38, 221: 19, 115: 24, 26: 14, 142: 9, 143: 46 };

/**
 * Prepare the global champion cache (en_us by default)
 * @param {String} lang 
 */
const getChampionData = async (lang = 'en_US') => {
    try {
        const fetch_result = await fetch(cupcake(`static/${lang}/championFull.json`, true, false));
        const json = await fetch_result.json();
        process.log.info(`Fetched champion ${json.version} data from CDN`);
        champion_en_US = json.data;
    } catch (err) {
        process.log.error('Failed to fetch champion data from CDN');
        process.log.error(err);
    }
}

/**
 * Ole knows what it does, i don´t.
 * @param {Object} input 
 * @param {Array} selectedGroups 
 * @returns 
 */
const queueGroupMerger = (input, selectedGroups) => {

    const data = deepClone(input);

    let groups = {};
    selectedGroups.forEach(group => groups[group] = true);

    if (selectedGroups.length === 0) {
        // we dont want 0 queue groups selected
        // that could lead to bugs, I dont like bugs
        return data;
    }

    let response = [];
    const queueGroups = data.data;

    for (let i = 0; i < queueGroups.length; i++) {
        const queueGroup = queueGroups[i];

        if (groups.hasOwnProperty(queueGroup.queueGroupId ?? 0)) {
            response.push(queueGroup);
        }

    }

    let pentaKillsIndexes = [];
    for (let i = 0; i < data.pentaKills.length; i++) {
        const pentakill = data.pentaKills[i];
        if (!groups.hasOwnProperty(pentakill.queueGroupId)) {
            pentaKillsIndexes.push(i);
        }
    }

    pentaKillsIndexes.sort((a, b) => a - b);
    for (let i = pentaKillsIndexes.length - 1; i >= 0; i--) {
        data.pentaKills.splice(pentaKillsIndexes[i], 1);
    }

    data.data = response;
    return data;
}

/**
 * Returns a champion by key based on the language
 * @param {String} key 
 * @param {String} lang
 * @returns 
 */
const getChampion = (key, lang = 'en_US') => {
    try {
        if (typeof key === "string") {
            if (!champion_en_US[key]) {
                process.log.error("Wrong type of key for: " + key);
                throw new Error("Wrong type of key for");
            }
            return champion_en_US[key];
        }

        key = key.toString();


        if (global_championCache[lang]) {
            if (global_championCache[lang][key]) {
                return global_championCache[lang][key];
            }
        } else {
            global_championCache[lang] = {};
        }

        const championList = Object.values(champion_en_US);
        for (let i = 0; i < championList.length; i++) {
            const champion = championList[i];
            if (champion.key === key) {
                global_championCache[lang][key] = champion;
                return champion;
            }
        }

    } catch (err) {
        process.log.error(err);
        process.exit(1);
    }

    process.log.error("Unkown champion: " + key);
    throw new Error("Unkown champion");
}

/**
 * Will calculate the rough size in bytes of the cached data
 * @returns {Number}
 */
const getCacheUsageInBytes = () => {
    if (process.env.CACHE_DRIVER === "FS") {
        return imageCache.roughSize();
    }
    const objectList = [];
    const stack = [...imageCache.entries()];
    const bytes = [0];
    while (stack.length) {
        const value = stack.pop();
        if (value == null) bytes[0] += 4;
        else if (typeof value === 'boolean') bytes[0] += 4;
        else if (typeof value === 'string') bytes[0] += value.length * 2;
        else if (typeof value === 'number') bytes[0] += 8;
        else if (typeof value === 'object' && objectList.indexOf(value) === -1) {
            objectList.push(value);
            if (value instanceof Map) {
                stack.push(...value.entries());
            } else if (typeof value.byteLength === 'number') {
                bytes[0] += value.byteLength;
            } else if (value[Symbol.iterator]) {
                for (const v of value) stack.push(v);
            } else {
                Object.keys(value).forEach(k => {
                    bytes[0] += k.length * 2; stack.push(value[k]);
                });
            }
        }
    }
    return bytes[0];
}


/**
 * Returns a skin id by champion id
 * @param {String} key 
 * @returns 
 */
const getSkinById = (key) => {
    return skinlist[key] ?? 0;
}

/**
 * Takes seconds and returns a string with days, hours, minutes and seconds
 * @param {Number} seconds 
 * @param {Boolean} noSeconds 
 * @param {Object} localization
 * @returns 
 */
const beautifySecondsToFixedString = (seconds, noSeconds = false, localization = { days: "d", hours: "h", minutes: "m", seconds: "s" }) => {

    let days = 0;
    let hours = 0;
    let minutes = 0;
    minutes = Math.floor(seconds / 60);
    seconds = seconds % 60;

    hours = Math.floor(minutes / 60);
    minutes = minutes % 60;

    days = Math.floor(hours / 24);
    hours = hours % 24;

    let result = '';
    if (days > 0) {
        result += `${days}${localization.days} `;
    }
    if (hours > 0 || days > 0) { // Include hours if there are days
        result += `${("0" + hours).slice(-2)}${localization.hours} `;
    }
    if (minutes > 0 || hours > 0 || days > 0) { // Include minutes if there are hours or days
        result += `${("0" + minutes).slice(-2)}${localization.minutes}`;
    }
    if (!noSeconds && (seconds > 0 || minutes > 0 || hours > 0 || days > 0)) { // Include seconds if there are minutes, hours, or days
        result += ` ${("0" + seconds).slice(-2)}${localization.seconds}`;
    }

    return result.trim();
}

/**
 * Download a webp image and return it as base64
 * @param {String} url 
 * @returns 
 */
const getWebPImageAsBase64 = async (url) => {

    // Check if the image for the given URL is already in the cache
    if (imageCache.has(url)) {
        served_bytes += imageCache.get(url).length;
        return await compressor.decompress(imageCache.get(url));
    }

    const response = await fetch(url);
    if (!response.ok) {
        process.log.error(`Failed to fetch the WebP image from ${url}.`);
        throw new Error('Failed to fetch the WebP image.');
    }

    // Get the buffer directly from the response
    const buffer = await response.arrayBuffer();
    const base64 = Buffer.from(buffer).toString('base64');
    const imageData = `data:image/webp;base64,${base64}`;

    // Store the fetched image data in the cache
    const compressedData = await compressor.compress(imageData);
    imageCache.set(url, compressedData);

    served_bytes += imageData.length;
    return imageData;
}

/**
 * Download a webp image (SummonerProfilePic) and return it as base64.
 * If fetching from the original ID fails, it will try fetching with the default ID of 29.
 * Images are cached after the first fetch to reduce network calls.
 * @param {Number} id - The ID to fetch the image for.
 * @returns {String} Base64 representation of the fetched webp image.
 */
const getProfileWebPImageAsBase64 = async (id) => {
    let usedFallback = id;
    // Check if the image for the given ID is already in the cache
    if (imageCache.has(id)) {
        served_bytes += imageCache.get(id).length;
        return await compressor.decompress(imageCache.get(id));
    }

    let response;

    try {
        response = await fetch(cupcake(`icons/${id}`, false));
        if (!response.ok) {
            throw new Error();
        }
    } catch (error) {
        process.log.warning(`Failed to fetch the WebP image for ID ${usedFallback}. Trying the default ID 29...`);
        usedFallback = "ProfilePicFallback";
        if (imageCache.has('ProfilePicFallback')) {
            served_bytes += imageCache.get(usedFallback).length;
            return await compressor.decompress(imageCache.get(usedFallback));
        }
        response = await fetch(cupcake(`icons/29`, false));
        if (!response.ok) {
            throw new Error('Failed to fetch the WebP image for both the given ID and the default ID 29.');
        }
    }

    const buffer = await response.arrayBuffer();
    const base64 = Buffer.from(buffer).toString('base64');
    const imageData = `data:image/webp;base64,${base64}`;

    // Store the fetched image data in the cache (Only if its not the fallback)
    const compressedData = await compressor.compress(imageData);
    imageCache.set(usedFallback, compressedData);

    served_bytes += imageData.length;
    return imageData;
}

/**
 * Will replace all ((Variables)) with the corresponding value of the Key
 * @param {String} templateid 
 * @param {Object} data 
 * @returns {String}
 */
function CSStemplate(templateid, data) {
    if (typeof (templateid) !== 'string' || typeof (data) !== 'object') {
        throw new Error('Template expects string and object');
    }
    return templateid.replace(
        /\(\((\w+)\)\)/g,
        function (m, key) {
            return data.hasOwnProperty(key) ? data[key] : "NULL";
        }
    );
}

/**
 * Will replace all {{Variables}} with the corresponding value of the Key
 * @param {String} templateid 
 * @param {Object} data 
 * @returns {String}
 */
const template = (templateid, data) => {
    if (typeof (templateid) !== 'string' || typeof (data) !== 'object') {
        throw new Error('Template expects string and object');
    }

    // Define a recursive function to handle nested replacements
    function replaceNestedPlaceholders(str) {
        if (!str.includes('{{')) return str;  // No more replacements needed

        return str.replace(/{{2}(\w*)}{2}/g, function (m, key) {
            if (data.hasOwnProperty(key)) {
                // If the value is a string that contains placeholders, replace them too
                if (typeof data[key] === 'string' && data[key].includes('{{')) {
                    return replaceNestedPlaceholders(data[key]);
                }
                return data[key];
            }
            return "0";
        });
    }

    // Start the replacement process
    return replaceNestedPlaceholders(templateid);
}

/**
 * Seperate a number using language-specific symbols
 * @param {Number} number 
 * @returns 
 */
const separateNumberByLocale = (number, langObj) => {
    const { lang_number_sep_thousand, lang_number_sep_decimal } = langObj
    const parts = number.toString().split('.');
    parts[0] = parts[0].replace(/\B(?=(\d{3})+(?!\d))/g, lang_number_sep_thousand);
    return parts.join(lang_number_sep_decimal);
};

/**
 * Shorten a number to its abbreviated form using language-specific symbols.
 * @param {number} num - The number to shorten.
 * @param {Object} langObj - The language object containing number abbreviations.
 * @returns {string} - The shortened number.
 */
function shortenNumberByLocale(num, langObj) {

    const scales = [
        { value: 1e15, symbol: langObj.number_quadrillion || 'Q' },
        { value: 1e12, symbol: langObj.number_trillion || 'T' },
        { value: 1e9, symbol: langObj.number_billion || 'B' },
        { value: 1e6, symbol: langObj.number_million || 'M' },
        { value: 1e3, symbol: langObj.number_thousand || 'K' }
    ];

    for (const scale of scales) {
        if (num >= scale.value) {
            return (num / scale.value).toFixed(1).replace(`.`, `${langObj.number_sep_decimal}`) + scale.symbol;
        }
    }
    return num;
}

/**
 * Get the champion data for a given champion ID.
 */
const clearimageCache = () => {
    imageCache.clear();
    process.log.info('Cleared image cache');
}

const getserved_bytes = () => {
    const served_mebibytes = served_bytes / 1024 / 1024;
    served_bytes = 0;
    return served_mebibytes;
}

module.exports = {
    getserved_bytes,
    queueGroupMerger,
    getCacheUsageInBytes,
    separateNumberByLocale,
    getProfileWebPImageAsBase64,
    getWebPImageAsBase64,
    shortenNumberByLocale,
    beautifySecondsToFixedString,
    getChampionData,
    getSkinById,
    getChampion,
    CSStemplate,
    template,
    clearimageCache
}