const ErrorStackParser = require('error-stack-parser');
const path = require('path');

const rootPath = path.join(__dirname, '..');

process.on('uncaughtException', function(err) {

    if(err.name === 'Error' && process.env.LOG_STACK == 'true') console.log(err);
    if(err.name === 'TypeError') console.log(err);
    if(err.name === 'ReferenceError') console.log(err);

    if(err.code === 'ECONNREFUSED') {
        process.log.error('Balancer refused connection, is it running?')
        process.eventbus.emit('reconnect');
        return;
    }

    if(process.env.LOG_STACK == 'true'){
        process.log.error(`${err.name} in ${err.path}:${err.fileline} - ${err.stack}`);
    } else {
        process.log.error(`${err.name} in ${err.path}:${err.fileline} - ${err.reason}`);
    }

    // If system error and not 0 exit with error code
    if( err.system && err.status !== 0) process.exit(err.status);
});

class CommunicationError extends Error {
    /**
     * @param {String} message | The message to display to the user
     */
    constructor(message) {
        super(message);
        const parsed = ErrorStackParser.parse(this);

        this.info = 'Communication with the load balancer failed.'
        this.reason = message;
        this.status = 0;
        this.system = true;
        this.path = path.relative(rootPath, parsed[0].fileName);
        this.fileline = parsed[0].lineNumber;
        this.name = 'CommunicationError';
    }
}

class GenericPuppeteerClusterError extends Error {
    /**
     * @param {String} message | The message to display to the user
     */
    constructor(message) {
        super(message);
        const parsed = ErrorStackParser.parse(this);

        this.info = 'Generic error with the puppeteer cluster.'
        this.reason = message;
        this.status = 0;
        this.system = true;
        this.path = path.relative(rootPath, parsed[0].fileName);
        this.fileline = parsed[0].lineNumber;
        this.name = 'GenericPuppeteerClusterError';
    }
}

class PuppeteerClusterError extends Error {
    /**
     * @param {String} message | The message to display to the user
     */
    constructor(message) {
        super(message);
        const parsed = ErrorStackParser.parse(this);

        this.info = 'Critical error with the puppeteer cluster.'
        this.reason = message;
        this.status = 1401;
        this.system = true;
        this.path = path.relative(rootPath, parsed[0].fileName);
        this.fileline = parsed[0].lineNumber;
        this.name = 'PuppeteerClusterError';
    }
}


module.exports = {
    CommunicationError,
    GenericPuppeteerClusterError,
    PuppeteerClusterError
}